// Copyright (c) Meta Platforms, Inc. and affiliates.
// All rights reserved.
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.

#pragma once

namespace facebook::torchcodec {

// See note in corresponding cpp file
bool loadNVCUVIDLibrary();

} // namespace facebook::torchcodec
